use chrono::NaiveDate;
use dimensioned::si;
use emseries::{DateTimeTz, Recordable, Timestamp};
use serde::{Deserialize, Serialize};

/// SetRep represents workouts like pushups or situps, which involve doing a "set" of a number of
/// actions, resting, and then doing another set.
pub struct SetRep {
    /// I assume that a set/rep workout is only done once in a day.
    date: NaiveDate,
    /// Each set entry represents the number of times that the action was performed in a set. So, a
    /// pushup workout that involved five sets would have five entries. Each entry would be x
    /// number of pushups. A viable workout would be something like [6, 6, 4, 4, 5].
    sets: Vec<u32>,
    comments: Option<String>,
}

/// The number of steps one takes in a single day.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct Steps {
    date: NaiveDate,
    count: u32,
}

/// TimeDistance represents workouts characterized by a duration and a distance travelled. These
/// sorts of workouts can occur many times a day, depending on how one records things. I might
/// record a single 30-km workout if I go on a long-distanec ride. Or I might record multiple 5km
/// workouts if I am out running errands. Distance and Duration are both optional because different
/// people have different priorities and may choose to measure different things.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct TimeDistance {
    /// The precise time (and the relevant timezone) of the workout. One of the edge cases that I
    /// account for is that a ride which occurred at 11pm in one timezone would then count as 1am
    /// if one moved two timezones to the east. This is kind of nonsensical from a human
    /// perspective, so the DateTimeTz keeps track of the precise time in UTC, but also the
    /// timezone in which the event was recorded.
    datetime: DateTimeTz,
    /// The distance travelled. This is optional because such a workout makes sense even without
    /// the distance.
    distance: Option<si::Meter<f64>>,
    /// The duration of the workout, which is also optional. Some people may keep track of the
    /// amount of distance travelled without tracking the duration.
    duration: Option<si::Second<f64>>,
    comments: Option<String>,
}

/// A singular daily weight measurement. Weight changes slowly enough that it seems unlikely to
/// need to track more than a single weight in a day.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct Weight {
    date: NaiveDate,
    weight: si::Kilogram<f64>,
}

/// The unified data structure for all records that are part of the app.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub enum TraxRecord {
    BikeRide(TimeDistance),
    Row(TimeDistance),
    Run(TimeDistance),
    Steps(Steps),
    Swim(TimeDistance),
    Walk(TimeDistance),
    Weight(Weight),
}

impl Recordable for TraxRecord {
    fn timestamp(&self) -> Timestamp {
        match self {
            TraxRecord::BikeRide(rec) => Timestamp::DateTime(rec.datetime.clone()),
            TraxRecord::Row(rec) => Timestamp::DateTime(rec.datetime.clone()),
            TraxRecord::Run(rec) => Timestamp::DateTime(rec.datetime.clone()),
            TraxRecord::Steps(rec) => Timestamp::Date(rec.date),
            TraxRecord::Swim(rec) => Timestamp::DateTime(rec.datetime.clone()),
            TraxRecord::Walk(rec) => Timestamp::DateTime(rec.datetime.clone()),
            TraxRecord::Weight(rec) => Timestamp::Date(rec.date),
        }
    }

    fn tags(&self) -> Vec<String> {
        vec![]
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use emseries::Series;

    #[test]
    fn can_record_records() {
        let file = tempfile::NamedTempFile::new().expect("a temporary file");
        let path = file.into_temp_path();
        let mut series: Series<TraxRecord> = Series::open(&path).unwrap();

        let record = TraxRecord::Steps(Steps {
            date: NaiveDate::from_ymd_opt(2023, 1, 1).unwrap(),
            count: 1000,
        });

        let id = series.put(record.clone()).unwrap();

        let record_ = series.get(&id).unwrap();
        assert_eq!(record_, record);
    }
}
